<* O2EXTENSIONS + *>
MODULE fDynamicParticle;

IMPORT
  Out, Part:=fParticle, Space:=fSpace, Thr:=SymThreads, SPACE:=Space,
  GPart:=GeneralParticle, Math:=mMath, PartMove:=fParticleMove, Area:=LocalArea,
  PHC:=PhConst, Vec:=m3DVectors, PList:=fParticleList, Polar:=mPolar,
  Sum:=MultiSum, CT:=CathodeTemperature, Rnd:=mRandom, COM:=Common, Protocol,
  Disp:=Display;

CONST
  ModuleName = 'fDynamicParticle';

  accomodation = 0.25;

TYPE
  Float = Math.Float;
  FloatArray2Ptr = Math.FloatArray2Ptr;
  Vector = Vec.Vector;
  IntArray2Ptr = Space.IntArray2Ptr;

  Field = POINTER TO RECORD
    Vz,Vr,SqrV,weight:FloatArray2Ptr;
    TotalWeight:Float;
    ReflWeight:FloatArray2Ptr;
  END;

  Fields = POINTER TO ARRAY OF Field; (* Fields in all the areas *)

	PropertiesDesc * = RECORD (Part.PropertiesDesc)
    new - : ARRAY Thr.Count OF Fields;
	END;
	Properties * = POINTER TO PropertiesDesc;

  Particle * = RECORD (GPart.Base)
    weight * : Float;
  END;

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE NewField (AreaIndex:LONGINT): Field;
  (* Returns a new field with indexes corresponding to the specified area,
     and filled by 0 *)
  VAR
    f:Field;
  BEGIN
    Disp.String(20,20,'',0);                     (* Another stupid invalid location in run mode is removed by this *)
    ASSERT(AreaIndex>=0);
    ASSERT(AreaIndex<LEN(SPACE.areas^));

    NEW(f);
    f.Vz:=Space.NewFloatArray(AreaIndex,0);
    f.Vr:=Space.NewFloatArray(AreaIndex,0);
    Disp.String(20,20,'',0);                     (* Another stupid invalid location in run mode is removed by this *)
    f.SqrV:=Space.NewFloatArray(AreaIndex,0);
    f.weight:=Space.NewFloatArray(AreaIndex,0);
    f.ReflWeight:=Space.NewFloatArray(AreaIndex,0);
    f.TotalWeight:=0;
    RETURN f
  END NewField;

PROCEDURE NewFields (): Fields;
  VAR
    f:Fields;
    i,AreaCount:LONGINT;
  BEGIN
    AreaCount:=LEN(SPACE.areas^);
    ASSERT(AreaCount>0);

    NEW(f,AreaCount);
    FOR i:=0 TO AreaCount-1 DO
      f[i]:=NewField(i);
    END;
    RETURN f
  END NewFields;

PROCEDURE Init * (p:Properties);
  (** Initiates the particle properties. *)
  VAR
    i:LONGINT;
  BEGIN
    Out.String('Start Init');
    ASSERT(p#NIL);
    Part.InitProperties(p);
    FOR i:=0 TO Thr.Count-1 DO
      p.new[i]:=NewFields();
    END;
    Out.String(' - End Init');
  END Init;

PROCEDURE SaveSpeeds (r-,v-:Vector; weight:Float; f:Field; iZ,iR:LONGINT);
  BEGIN
    ASSERT(weight>=0);
    ASSERT(f#NIL);
    ASSERT(iZ>=0);
    ASSERT(iZ<LEN(f.Vz^));
    ASSERT(iR>=0);
    ASSERT(iR<LEN(f.Vz^,1));

    f.Vz[iZ,iR]:=f.Vz[iZ,iR]+weight*v.z;
    IF (r.x#0) OR (r.y#0) THEN
      f.Vr[iZ,iR]:=f.Vr[iZ,iR]+weight*Polar.Vr(r.x,r.y,v.x,v.y);
    END;
    f.SqrV[iZ,iR]:=f.SqrV[iZ,iR]+weight*Vec.SqrAbs(v);
  END SaveSpeeds;

PROCEDURE SaveMacroParameters (p:Properties; AreaIndex:LONGINT; rv-:Particle);
  VAR
    z,r,weight:Float;
    a:Area.Area;
    iZ,iR,ThrInd:LONGINT;
    f:Field;
  BEGIN
    weight:=rv.weight;
    ASSERT(weight>=0);
    ASSERT(p#NIL);
    ASSERT(AreaIndex>=0);
    ASSERT(AreaIndex<LEN(SPACE.areas^));

    a:=SPACE.areas[AreaIndex];
    z:=rv.r.z;
    r:=Vec.R(rv.r);
    ASSERT(a.Inside(z,r));
    iZ:=Area.ZToInd(a,z);
    iR:=Area.RToInd(a,r);
    ASSERT(iZ>=0);
    ASSERT(iZ<=Area.LastZ(a));
    ASSERT(iR>=0);
    ASSERT(iR<=Area.LastR(a));

    ThrInd:=Thr.Index();
    f:=p.new[ThrInd][AreaIndex];

    SaveSpeeds(rv.r,rv.v,weight,f,iZ,iR);

    f.weight[iZ,iR]:=f.weight[iZ,iR]+weight;
  END SaveMacroParameters;

PROCEDURE Move * (p:Properties; VAR rv:Particle; time:Float): BOOLEAN;
  (** Moves preliminary initialized particle rv of type p during the specified
      time, and fills corresponding records of p. Returns if the movement was
      successful *)
  VAR
    valid:BOOLEAN;
    ThrInd,AreaIndex:LONGINT;
    f:Field;
  BEGIN
    ASSERT(p#NIL);
    ASSERT(time>=0);

    ThrInd:=Thr.Index();
    AreaIndex:=SPACE.FindAreaIndexByR(rv.r);
    ASSERT(AreaIndex>=0);
    f:=p.new[ThrInd][AreaIndex];
    f.TotalWeight:=f.TotalWeight+rv.weight;

    valid:=PartMove.InSpace(p,rv,time,PartMove.inTime);
    IF valid THEN
      AreaIndex:=SPACE.FindAreaIndexByR(rv.r);
      IF AreaIndex>=0 THEN
        ASSERT(AreaIndex<LEN(SPACE.areas^));
        SaveMacroParameters(p,AreaIndex,rv);
      END;
    END;
    RETURN valid
  END Move;

VAR
  FinalizeData:RECORD
    p:Properties;
    NumberOfRealParticles:Float;
    TotalWeight:Sum.Sum;
    weight:Float;
  END;

PROCEDURE GetAndClearTotalWeightInThreads;
  VAR
    p:Properties;
    AreaIndex,ThrInd:LONGINT;
    nf:Field;
  BEGIN
    p:=FinalizeData.p;
    ASSERT(p#NIL);

    ThrInd:=Thr.Index();
    FOR AreaIndex:=0 TO LEN(SPACE.areas^)-1 DO
      nf:=p.new[ThrInd][AreaIndex];
      Sum.Add(FinalizeData.TotalWeight,nf.TotalWeight);
      nf.TotalWeight:=0;
    END;
  END GetAndClearTotalWeightInThreads;

PROCEDURE GetAndClearTotalWeight(): Float;
  BEGIN
    ASSERT(~Thr.executing);
    ASSERT(FinalizeData.p#NIL);
    Sum.Clear(FinalizeData.TotalWeight);
    Thr.Exec(GetAndClearTotalWeightInThreads);
    RETURN Sum.Result(FinalizeData.TotalWeight);
  END GetAndClearTotalWeight;

PROCEDURE FinalizeInThreads;
  (** Finalizes the given properties after calculations. The final concentration
      is multiplied by the specified factor. *)
  VAR
    AreaIndex,iZ,iR,IndexOfThread,LastZ,LastR,ThrInd:LONGINT;
    of,nf:Field;
    field:Space.Field;
    a:Area.Area;
    Vz,Vr,SqrV,weight,ReflWeight,w:Float;
    TotalWeight:Float;
    p:Properties;
    NumberOfRealParticles:Float;
  BEGIN
    p:=FinalizeData.p;
    NumberOfRealParticles:=FinalizeData.NumberOfRealParticles;

    ASSERT(p#NIL);
    ASSERT(NumberOfRealParticles>0);

    TotalWeight:=FinalizeData.weight;
    ASSERT(TotalWeight>0);

    ThrInd:=Thr.Index();

    FOR AreaIndex:=0 TO LEN(SPACE.areas^)-1 DO
      a:=SPACE.areas[AreaIndex];
      LastZ:=Area.LastZ(a);
      LastR:=Area.LastR(a);
      nf:=p.new[0][AreaIndex];

      FOR IndexOfThread:=1 TO Thr.Count-1 DO
        of:=p.new[IndexOfThread][AreaIndex];
        FOR iZ:=ThrInd TO LastZ BY Thr.Count DO
          FOR iR:=0 TO LastR DO
            nf.Vz[iZ,iR]:=nf.Vz[iZ,iR]+of.Vz[iZ,iR];
            nf.Vr[iZ,iR]:=nf.Vr[iZ,iR]+of.Vr[iZ,iR];
            nf.SqrV[iZ,iR]:=nf.SqrV[iZ,iR]+of.SqrV[iZ,iR];
            nf.weight[iZ,iR]:=nf.weight[iZ,iR]+of.weight[iZ,iR];
            nf.ReflWeight[iZ,iR]:=nf.ReflWeight[iZ,iR]+of.ReflWeight[iZ,iR];
            of.Vz[iZ,iR]:=0;
            of.Vr[iZ,iR]:=0;
            of.SqrV[iZ,iR]:=0;
            of.weight[iZ,iR]:=0;
            of.ReflWeight[iZ,iR]:=0;
          END;
        END;
      END;

      field:=p.old[AreaIndex];

      FOR iZ:=ThrInd TO LastZ BY Thr.Count DO
        FOR iR:=0 TO LastR DO
          weight:=nf.weight[iZ,iR];
          ASSERT(weight>=0);
          IF weight=0 THEN
            field.Vz[iZ,iR]:=0;
            field.Vr[iZ,iR]:=0;
            field.concentration[iZ,iR]:=0;
            field.temperature[iZ,iR]:=0;
          ELSE
            ASSERT(weight>0);

            field.concentration[iZ,iR]:= weight/Area.ElVol(a,iZ,iR)
                                        *NumberOfRealParticles/TotalWeight;

            ReflWeight:=nf.ReflWeight[iZ,iR];
            ASSERT(ReflWeight>=0);
            w:=weight+ReflWeight;
            ASSERT(w>0);

            Vz:=nf.Vz[iZ,iR]/w; (* Mean Vz *)
            Vr:=nf.Vr[iZ,iR]/w; (* Mean Vr *)
            SqrV:=nf.SqrV[iZ,iR]/w - (sqr(Vz)+sqr(Vr));
            (* SqrV = Mean(sqr(v[i]))-sqr(Mean(v[i])) *)
            ASSERT(SqrV>=0);

            field.temperature[iZ,iR]:=SqrV*p.mass/(3*PHC.k);
            field.Vz[iZ,iR]:=Vz;
            field.Vr[iZ,iR]:=Vr;
          END;

          nf.Vz[iZ,iR]:=0;
          nf.Vr[iZ,iR]:=0;
          nf.SqrV[iZ,iR]:=0;
          nf.weight[iZ,iR]:=0;
          nf.ReflWeight[iZ,iR]:=0;
        END;
      END;

    END;
  END FinalizeInThreads;

PROCEDURE Finalize * (p:Properties; NumberOfRealParticles:Float);
  BEGIN
    ASSERT(~Thr.executing);
    ASSERT(p#NIL);
    FinalizeData.p:=p;
    FinalizeData.NumberOfRealParticles:=NumberOfRealParticles;
    FinalizeData.weight:=GetAndClearTotalWeight();
    ASSERT(FinalizeData.weight>0);
    Thr.Exec(FinalizeInThreads);
    FinalizeData.p:=NIL;
  END Finalize;

PROCEDURE CalculateReflectedStream (p:Properties; msg-:PartMove.CollisionMsg; VAR rv:Particle);
  VAR
    ReflV:Vector;
    iZ,iR,ThrInd,AreaIndex:LONGINT;
    f:Field;
    ap:Properties;
  BEGIN
    ASSERT(p#NIL);
    ASSERT(msg.AreaIndex>=0);
    ASSERT(msg.AreaIndex<LEN(SPACE.areas^));
    ASSERT(SPACE.FindAreaIndexByR(rv.r)=msg.AreaIndex);

    AreaIndex:=msg.AreaIndex;
    ASSERT(msg.iZ>=0);
    ASSERT(msg.iZ<=Area.LastZ(SPACE.areas[AreaIndex]));
    ASSERT(msg.iR>=0);
    ASSERT(msg.iR<=Area.LastR(SPACE.areas[AreaIndex]));

    iZ:=msg.iZ;
    iR:=msg.iR;
    ThrInd:=Thr.Index();
    f:=p.new[ThrInd][AreaIndex];

    ASSERT(msg.ap#NIL);
    ap:=msg.ap(Properties);
    ASSERT(ap.mass>0);
    ASSERT(p.mass>0);

    ReflV.x := (ap.mass*msg.av.x + p.mass*msg.tv.x - ap.mass*rv.v.x)/p.mass;
    ReflV.y := (ap.mass*msg.av.y + p.mass*msg.tv.y - ap.mass*rv.v.y)/p.mass;
    ReflV.z := (ap.mass*msg.av.z + p.mass*msg.tv.z - ap.mass*rv.v.z)/p.mass;

    SaveSpeeds(rv.r,ReflV,rv.weight,f,iZ,iR);
    f.ReflWeight[iZ,iR]:=f.ReflWeight[iZ,iR]+rv.weight;
  END CalculateReflectedStream;

PROCEDURE CorrectReflectedSpeed (p:Properties; msg-:PartMove.ReflectionMsg;
                                 VAR rv:Particle);
  VAR
    n:Vector;
    T,SqrV:Float;
    ThrInd:LONGINT;
  BEGIN
    ASSERT(p#NIL);
    ASSERT(msg.w#NIL);

    ThrInd:=Thr.Index();
    IF Rnd.Real(COM.rnd[ThrInd])<accomodation THEN
      msg.w.Normal(rv.r.x,rv.r.y,n);
      IF Vec.ScalProd(rv.v,n)<0 THEN
        n.x:=-n.x;
        n.y:=-n.y;
        n.z:=-n.z;
      END;

      T:=msg.w.Temperature(rv.r);

      IF CT.Initialized() THEN
        SqrV:=Vec.SqrAbs(rv.v);
        GPart.InitSpeedAtWall(rv.v,p.mass,T,n);
        SqrV:=Vec.SqrAbs(rv.v)-SqrV;
        CT.AddEnergy(Vec.R(rv.r), -rv.weight*p.mass*SqrV/2 );
      ELSE
        GPart.InitSpeedAtWall(rv.v,p.mass,T,n);
      END;
    END;
  END CorrectReflectedSpeed;

PROCEDURE (p:Properties) HandleMsgWithParticle * (VAR msg:Part.Msg; VAR rv:GPart.Base);
  BEGIN
    ASSERT(rv IS Particle);
    WITH msg:PartMove.CollisionMsg DO
      CalculateReflectedStream (p,msg,rv(Particle))
    |msg:PartMove.ReflectionMsg DO
      CorrectReflectedSpeed(p,msg,rv(Particle))
    ELSE
    END;
  END HandleMsgWithParticle;

PROCEDURE (p:Properties) HandleMsg * (VAR msg:Part.Msg);
  BEGIN
    WITH msg:PList.ProtocolMsg DO
      Part.ProtocolProperties(p);
    ELSE
    END;
  END HandleMsg;

BEGIN
  Protocol.String('MODULE ',ModuleName,':');
  Protocol.Real('accomodation=',accomodation,'');
  Protocol.String('','','');
END fDynamicParticle.